/*      
  ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/rdproj.cpp $
 $LastChangedRevision: 91 $
 $Date: 2014-07-02 17:34:27 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 RDPROJ GENERATES THE RANDOM ACCESS VERSION OF THE PROJECTION DATA
 ON  PRJFIL.
 THE MODIFIER ON  PROJECTION CARD IS ONE OF THE FOLLOWING -
 REAL     THIS IS THE DEFAULT VALUE.  PROJECTION DATA ARE READ FROM
 FILE11  AND PUT ON PRJFIL
 PSEUDO   CREATES PSEUDO PROJECTION DATA FROM INFORMATION ON FILE11
 AND INPUT  AND PUTS ON PRJFIL.  THIS REQUIRES THAT
 PICTURE  TEST  BE ALREADY SPECIFIED.
 OLD      THIS ASSUMES THAT PRJFIL ALREADY EXISTS FROM A PREVIOUS
 SNARK RUN.  THE GEOMETRY AND NOISE INFORMATION ARE IN THE
 FIRST RECORD OF  PRJFIL.

 SEE SNARK MANUAL FOR MORE DETAILS.
 *
 *
 *Modification date:  July 13, 2008
 *Author: Joanna Klukowska (jk)
 *Changes:  Added functionality for beam hardening correction.
 *          If the user sets the options for correction in the input
 *          file, rdproj() createas nergy temporary prjfil's with to 
 *          pass to bh_correction::correction() for beam hardening correction
 *          which is responsible for creation of the final prjfil
 *          that is used in the remainder of the program.
 *          The option of beam hardening correction is only appropriate 
 *          when polychromatic data is provided.
 *
 */

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#include "blkdta.h"
#include "creacm.h"
#include "geom.h"
#include "anglst.h"
#include "spctrm.h"
#include "consts.h"
#include "uiod.h"

#include "raylen.h"
#include "posit.h"
#include "DIGGauss.h"
#include "creaer.h"
#include "bldlst.h"
#include "pseudo.h"
#include "bh_correction.h"

#include "recfile.h"
#include "file11.h"
#include "projfile.h"
#include "infile.h"

#include "rdproj.h"
#include "bh_correction.h"

#include <time.h>

void rdproj()
{

	static const INTEGER proj_codes[4] =
	{ CHAR2INT('o', 'l', 'd', ' '), CHAR2INT('r', 'e', 'a', 'l'), CHAR2INT('p',
			's', 'e', 'u'), CHAR2INT('b', 'h', 'p', 's')    //jk 9/19/2008
			//used only for recursive runs of snark during the
			//beam hardening correction process
			//should not be listed as an option in the manual
			};

	// jk 7/13/2008 added for beam hardening correction
	static const INTEGER bhc_codes[1] =
	{ CHAR2INT('b', 'e', 'a', 'm') };

	static BOOLEAN called = FALSE;

	REAL dtor;
	INTEGER nword;
	INTEGER nr;
	REAL umean;
	REAL amean;
	REAL totden;
	REAL totlen;
	REAL hfside;
	INTEGER np;
	REAL sinth;
	REAL theta;
	REAL costh;
	REAL ang;
	REAL the;
	REAL angle;
	INTEGER numb;
	REAL snorm;

	INTEGER* lbase = NULL;
	REAL* wbase = NULL; //wei 3/2005

	int i, j, k;

	BOOLEAN eol;
	REAL ax, ay, mx, my;

	// jk 7/13/2008 added for beam hardening correction

	INTEGER bhcword; //BHC keyword read from the input file
	SnarkPrjFile * BHC_PrjFile = NULL; //temporary prjfil's
	char filename[256];
	char command[256];
	bh_correction bh_cor;

	///////////////////////////////////////////////////////////

	// CHECK FOR MULTIPLE CALLS

	if (called)
		creaer(1);

	called = TRUE;
	dtor = (REAL) (Consts.pi / 180.0);

	// CHECK THAT RDPICT HAS BEEN PERFORMED BEFORE THIS

	if (GeoPar.nelem <= 0)
		creaer(23);

	// DETERMINE THE PROJECTION MODIFIER AND THE INPUT FILE

	// changed last parameter from "4" to "3". lajos, Nov 18, 2004
	nword = InFile.getwrd(FALSE, &eol, proj_codes, 4);

	// IF OLD  IS OPTED JUST READ HEADER & QUIT

	if (nword == proj_codes[0]) // OLD
	{
		// OPEN PROJECTION FILE FOR PROJECTION OLD CASE

		if (ProjFile.Open((char *) "prjfil") != 0)
		{
			fprintf(output, "\n **** unable to open prjfil");
			fprintf(output, "\n **** program aborted\n");
			exit(-1);
		};

		ProjFile.ReadHeader(Creacm.name, &NoisePar, &Spctrm, &Anglst);

		// compute average density - fixes bug 232 of snark05
		double *projdata = new double[GeoPar.nrays];
		totden = 0.0;
		totlen = 0.0;
		hfside = (REAL) 0.5 * GeoPar.nelem * GeoPar.pixsiz;
		for (np = 0; np < GeoPar.prjnum; ++np)
		{
			ProjFile.ReadProj(np, projdata, GeoPar.nrays);
			for (nr = 0; nr < GeoPar.nrays; nr++)
			{
				if (GeoPar.strip)
				{
					if (GeoPar.vri)
					{
						totden +=
								(REAL) (projdata[nr]
										/ (GeoPar.pinc
												* MAX0(fabs(sinth), fabs(costh))));
					}
					else
					{
						totden += projdata[nr] / GeoPar.pinc;
					}
				}
				else
				{
					totden += projdata[nr];
				}
				posit(np, nr, &ax, &ay, &mx, &my);
				totlen += raylen(SOT_rect, ax, ay, mx, my, 0.0, 0.0, hfside,
						hfside, 1.0, 0.0);
			}
		}
		GeoPar.aveden = totden / totlen;
		fprintf(output, "\n         estimate of totlen = %16.6f", totlen);
		fprintf(output, "\n         estimate of totden = %16.6f", totden);
		fprintf(output, "\n         estimate of average density = %10.4f",
				GeoPar.aveden);
		delete[] projdata;
	}

	else
	{
		// READ PREFACE TO PROJECTION DATA

		if (nword == proj_codes[2] || nword == proj_codes[3]) // PSEUDO
		{
			NoisePar.resetNoise(); // jklukowska, bug 263, in case the noise was set
								   // previously, we need to reset it to initial values
			InFile.CreaInProj();
		}
		else
		{
			if (nword == proj_codes[1]) // REAL
			{
				if (!File11.isOpen)
				{
					if (File11.Open(FALSE) != 0)
					{
						fprintf(output,
								"\n **** unable to open file11 for reading");
						fprintf(output, "\n **** program aborted\n");
						exit(-1);
					};
					File11.isOpen = TRUE;
				};

				File11.CreaInProj();
			}
			else
			{
				fprintf(output,
						"\n **** OLD, REAL or PSEUDO must be specified");
				fprintf(output, "\n **** program aborted\n");
				exit(-1);
			}
		};

		if (Creacm.erflag)
			exit(555);

		// jk 7/13/2008 added for beam hardening correction
		//check if bh_correction flag is set
		bhcword = InFile.getwrd(FALSE, &eol, bhc_codes, 1);
		if (bhcword == bhc_codes[0])
		{        //beam hardening correction flag set
			bh_cor.bhc = TRUE;
			//check if data is polychromatic
			if (Spctrm.nergy == 1) //bhc cannot be performed
			{
				fprintf(output,
						"\n **** Beam hardening correction is not appropriate\n"
								"for monochromatic data (nergy = 1).\n"
								"Program will continue ignoring BHC command.\n\n");
				bh_cor.bhc = FALSE;
			}
			else
			{

				//read in the parameters for beam hardening correction
				bh_cor.readInputFilePoly();

				//allocate temporary prjfil's
				BHC_PrjFile = new SnarkPrjFile[1];

				time_t t;
				time(&t);   //get the current time
				//create temporary directory
				sprintf(bh_cor.tempDir, "bhc%i", (int) t);
				sprintf(command, "mkdir %s", bh_cor.tempDir);
				if (system(command) != 0)
				{
					fprintf(stdout, "\nERROR creating temporary directory %s\n",
							filename);
					exit(1);
				}
			}
		}

		if (ProjFile.Open((char*) "prjfil", GeoPar.usrays, GeoPar.prjnum) != 0)
		{
			fprintf(output, "\n **** unable to open prjfil");
			fprintf(output, "\n **** program aborted\n");
			exit(-1);
		};

		// jk 7/13/2008 added for beam hardening correction
		//open temporary prjfil's
		if (bh_cor.bhc)
		{
			for (i = 0; i < 1; i++)
			{
				sprintf(filename, "%s/q_p0", bh_cor.tempDir);
				if (BHC_PrjFile[i].Open(filename, GeoPar.usrays, GeoPar.prjnum)
						!= 0)
				{
					fprintf(output, "\n **** unable to open temporary prjfil");
					fprintf(output, "\n **** program aborted\n");
					exit(-1);
				}
			}
		}

		ProjFile.WriteHeader(Creacm.name, &GeoPar, &NoisePar, &Spctrm);

		ProjFile.WriteAngles(Anglst.bth);

		// jk 7/13/2008 added for beam hardening correction
		// write header info to temporary prjfil's
		if (bh_cor.bhc)
		{
			for (i = 0; i < 1; i++)
			{
				BHC_PrjFile[i].WriteHeader(Creacm.name, &GeoPar, &NoisePar,
						&Spctrm);
				BHC_PrjFile[i].WriteAngles(Anglst.bth);
			}
		}

		// ALLOCATE SPACE FOR ARRAYS LIST,WEIGHT,SPHI,CPHI,PBASE

		bldlst(&lbase, &wbase);

		if (GeoPar.div)
		{
			Anglst.genphi();
		}

		Anglst.pbase = new REAL[GeoPar.nrays];

		for (nr = GeoPar.fusray; nr <= GeoPar.lusray; nr++)
		{
			Anglst.pbase[nr] = 0.0;
		}

		// IF PROJECTION DATA ARE NOISY WITH BIAS DUE TO SPECIFICATION OF
		// SKEWED DISTRIBUTIONS PREPARE HERE TO REMOVE BIAS

		umean = 1.0;

		if (NoisePar.ultnmn != 0.0)
		{
			umean = (REAL) 1.0 / NoisePar.ultnmn;
		}

		amean = 0.0;

		if (NoisePar.addnfl)
		{
			amean = NoisePar.addnmn;
		}

		// INITIALIZE FOR COMPUTATION OF AVERAGE DENSITY

		totden = 0.0;
		totlen = 0.0;
		hfside = (REAL) 0.5 * GeoPar.nelem * GeoPar.pixsiz;

		// CHECK TO SEE IF PHANTOM IS PRESENT WHEN PSEUDO IS OPTED

		if ((nword == proj_codes[2] || nword == proj_codes[3])
				&& (Creacm.test == NULL))
		{
			creaer(24);
		}

		// SELECT PROJECTION BY PROJECTION AND PREPARE  PRJFIL

		for (np = 0; np < GeoPar.prjnum; np++)
		{
			if (nword == proj_codes[1])
			{

				if (!File11.ReadProjection(&theta, Anglst.pbase + GeoPar.fusray,
						GeoPar.lusray - GeoPar.fusray + 1))
				{
					fprintf(output,
							"\n **** end of file encountered on file11");
					fprintf(output, "\n **** program aborted\n");
					if (lbase != NULL)
						delete[] lbase;
					if (wbase != NULL)
						delete[] wbase; //wei 3/2005
					exit(-1);
				}

				sinth = (REAL) sin(theta);
				costh = (REAL) cos(theta);
				ang = theta / dtor;

				the = *(Anglst.bth + np);

				angle = the / dtor;
				if ((REAL) fabs(theta - the) > (REAL) 0.001)
				{
					fprintf(output, "\n **** angles specified for projection "
							"%4i do not match", np);
					fprintf(output, "\n      angle given in header is %10.4f "
							"degrees = %10.4f radians", angle, the);
					fprintf(output, "\n      angle given in data   is %10.4f "
							"degrees = %10.4f radians", ang, theta);
					fprintf(output, "\n **** program aborted\n");
					if (lbase != NULL)
						delete[] lbase;
					if (wbase != NULL)
						delete[] wbase; //wei 3/2005
					exit(-1);
				}
			}
			else
			{
				// PROJECTION  PSEUDO or BHPSEUDO  OPTION

				theta = Creacm.pang[np];

				sinth = (REAL) sin(theta);
				costh = (REAL) cos(theta);

				// bug221 start - swr - 03/03/07
				for (nr = 0; nr < GeoPar.nrays; nr++)
					Anglst.pbase[nr] = 0.0;

				for (nr = GeoPar.fusray; nr <= GeoPar.lusray; nr++)
				{
					REAL temp;
					if (nword == proj_codes[3])
					{
						//BHPSEUDO option
						//jk 07/20/2008 modyfying how pseudo projections are
						// computed - they need to be polychromatic for
						//beam hardening correction
						temp = 0.0;
						for (j = 0; j < Spctrm.nergy; j++)
						{
							temp += Spctrm.engwt[j]
									* NoisePar.raysum(
											pseudo(phantom_poly_array[j], np,
													nr, lbase, wbase, &numb,
													&snorm, GeoPar.line,
													FALSE));
						}
					}
					else
					{
						//PSEUDO option
						temp = NoisePar.raysum(
								pseudo(Creacm.test, np, nr, lbase, wbase, &numb,
										&snorm, GeoPar.line, FALSE));
					}
					if (NoisePar.quanin > 0 && NoisePar.quanin != 4)
						temp *= NoisePar.quanmn;
					Anglst.pbase[nr] = NoisePar.applyQuantumNoise(np, nr, temp);
				}

				// TAKE CARE OF NOISE

				NoisePar.scatterNoise(Anglst.pbase);

				for (nr = GeoPar.fusray; nr <= GeoPar.lusray; nr++)
				{
					REAL temp = NoisePar.computeAttenuation(np, nr,
							Anglst.pbase[nr]);
					temp = NoisePar.multiplicativeNoise(temp);
					Anglst.pbase[nr] = NoisePar.additiveNoise(temp);
				}
				// bug221 end - swr - 03/03/07
			}
			// REMOVE BIAS IN PROJECTION DATA POSSIBLY INTRODUCED BY A SKEWED
			// NOISE DISTRIBUTION

			if (NoisePar.ultnfl || NoisePar.addnfl)
			{
				for (i = GeoPar.fusray; i <= GeoPar.lusray; i++)
				{
					Anglst.pbase[i] = (Anglst.pbase[i] - amean) * umean;
				}
			}

			// ADD UP FOR ESTIMATION OF AVERAGE DENSITY FROM RAYSUMS

			for (nr = GeoPar.fusray; nr <= GeoPar.lusray; nr++)
			{
				// Changed here for SNARK89 Version 1.01 7/13/90 to conform to manual
				if (GeoPar.strip)
				{
					if (GeoPar.vri)
					{
						totden +=
								(REAL) (Anglst.pbase[nr]
										/ (GeoPar.pinc
												* MAX0(fabs(sinth), fabs(costh))));
					}
					else
					{
						totden += Anglst.pbase[nr] / GeoPar.pinc;
					}
				}
				else
				{
					totden += Anglst.pbase[nr];
				}

				posit(np, nr, &ax, &ay, &mx, &my);

				totlen += raylen(SOT_rect, ax, ay, mx, my, 0.0, 0.0, hfside,
						hfside, 1.0, 0.0);

			}

			// jk 7/13/2008 added for beam hardening correction
			// write data to temporary prjfil's if bhc is set
			if (bh_cor.bhc)
			{
				for (i = 0; i < 1; i++)
				{

					//iterate over values of Anglst.pbase and multiply by
					//coefficients of the correction polynomial
					//for (nr = GeoPar.fusray; nr <= GeoPar.lusray; nr++)
					INTEGER NoOfRays = GeoPar.lusray - GeoPar.fusray + 1;
					REAL t, tt;
					REAL * temp = new REAL[NoOfRays];
					for (k = 0; k < NoOfRays; k++)
						temp[k] = 0;
					for (nr = 0; nr < NoOfRays; nr++)
					{
						t = Anglst.pbase[nr + GeoPar.fusray];
						tt = 1;
						for (j = 0; j <= bh_cor.noDegree; j++)
						{
							temp[nr] += bh_cor.a[j] * tt;
							tt = tt * t;
						}

					}
					BHC_PrjFile[i].WriteProj(np, temp); //&(Anglst.pbase[GeoPar.fusray]));
					delete[] temp;
				}
			}
			else //write data to the final prjfil
			{

				// WRITE ON PRJFIL.
				ProjFile.WriteProj(np, &(Anglst.pbase[GeoPar.fusray]));
			}
		}
		if (!bh_cor.bhc)
		{
			GeoPar.aveden = totden / totlen;
			fprintf(output, "\n\n         estimate of totlen = %16.6f", totlen);

			fprintf(output, "\n         estimate of totden = %16.6f", totden);

			fprintf(output, "\n         estimate of average density = %10.4f",
					GeoPar.aveden);
		}
		// WRITE UPDATED HEADER ON PRJFIL

		// jk 7/13/2008 added for beam hardening correction
		// write header to temporary prjfil's if bhc is set
		if (bh_cor.bhc)
		{
			for (i = 0; i < 1; i++)
			{
				BHC_PrjFile[i].WriteHeader(Creacm.name, &GeoPar, &NoisePar,
						&Spctrm);
			}
		}
		else //write header to the final prjfil
		{

			ProjFile.WriteHeader(Creacm.name, &GeoPar, &NoisePar, &Spctrm);
		}

		// PREPARE TO EXIT FROM ROUTINE

		//free(Anglst.bth); deleted now in the destructor
	}

	// CLOSE THE PRJFIL

	// jk 7/13/2008 added for beam hardening correction
	if (bh_cor.bhc)
	{
		for (i = 0; i < 1; i++)
		{
			BHC_PrjFile[i].Close();
		}
	}
	else
	{

		ProjFile.Close();
	}

	// jk 7/13/2008 added for beam hardening correction
	// CALL BEAM HARDENING CORRECTION ROUTINES
	if (bh_cor.bhc)
	{
		ProjFile.Close();
		bh_cor.correction(BHC_PrjFile);

		// OPEN PROJECTION FILE JUST CORRECTED FOR BEAM HARDENING

		if (ProjFile.Open((char*) "prjfil") != 0)
		{
			fprintf(output,
					"\n **** unable to open prjfil after beam hardening");
			fprintf(output, "\n **** program aborted\n");
			exit(-1);
		};

		ProjFile.ReadHeader(Creacm.name, &NoisePar, &Spctrm, &Anglst);

		// compute average density
		double *projdata = new double[GeoPar.nrays];
		totden = 0.0;
		totlen = 0.0;
		hfside = (REAL) 0.5 * GeoPar.nelem * GeoPar.pixsiz;
		for (np = 0; np < GeoPar.prjnum; ++np)
		{
			ProjFile.ReadProj(np, projdata, GeoPar.nrays);
			for (nr = 0; nr < GeoPar.nrays; nr++)
			{
				if (GeoPar.strip)
				{
					if (GeoPar.vri)
					{
						totden +=
								(REAL) (projdata[nr]
										/ (GeoPar.pinc
												* MAX0(fabs(sinth), fabs(costh))));
					}
					else
					{
						totden += projdata[nr] / GeoPar.pinc;
					}
				}
				else
				{
					totden += projdata[nr];
				}
				posit(np, nr, &ax, &ay, &mx, &my);
				totlen += raylen(SOT_rect, ax, ay, mx, my, 0.0, 0.0, hfside,
						hfside, 1.0, 0.0);
			}
		}
		GeoPar.aveden = totden / totlen;
		fprintf(output, "\n         estimate of totlen = %16.6f", totlen);
		fprintf(output, "\n         estimate of totden = %16.6f", totden);
		fprintf(output, "\n         estimate of average density = %10.4f",
				GeoPar.aveden);
		delete[] projdata;

		ProjFile.Close();
	}
	if (!bh_cor.bhc)
	{
		fprintf(output, "\n         projection data read");
		fprintf(output, "\n         %s", Creacm.name);
	}
	// WRITE PROJECTION HEADER INTO RECORD nelem + 3 OF
	// RECONSTRUCTION FILE

	RecFile.SetProjName(Creacm.name);

	//delete the created directory bhc
	if (bh_cor.bhc)
	{
		sprintf(command, "rm -rf %s", bh_cor.tempDir);
		if (system(command) != 0)
		{
			fprintf(stdout, "\n\n****ERROR deleting temporary directory %s\n\n",
					bh_cor.tempDir);
		}
	}

	if (lbase != NULL)
		delete[] lbase;
	if (wbase != NULL)
		delete[] wbase; //wei 3/2005
	if (BHC_PrjFile != NULL)
		delete[] BHC_PrjFile;

	for (i = 6; i >= 0; i--)
	{
		if (phantom_poly_array[i] != NULL)
			delete[] phantom_poly_array[i];
	}

	return;
}
