/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/rfl.cpp $
 $LastChangedRevision: 94 $
 $Date: 2014-07-02 18:53:30 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 RHO-FILTERED LAYERGRAM RECONSTRUCTION METHOD
 THE CONTROL CARD FOLLOWING THE EXECUTE CARD CONTAINS (IN THE
 SNARK FREE FORMAT) THE FOLLOWING INFORMATION:
 1) THE DESIRED FILTER ("BAND", "SINC", "COSINE" OR "HAMMING")
 2) THE CUTOFF PARAMETER (OR THE VALUE OF ALPHA FOR HAMMING)
 3) THE INTERPOLATION METHOD DESIRED
 4) BACK PROJECTION REGION SIZE (OPTIONAL)
 THE FILTERS ARE CIRCULARLY SYMMETRIC VERSIONS OF THE
 CORRESPONDING FILTERS USED IN THE CONVOLUTION ALGORITHM.
 THE BAND LIMITING FILTER CORRESPONDS TO THE FILTER USED BY
 BRACEWELL + RIDDLE[1] AND RAMACHANDRAN + LAKSHMINARAYANAN[2].
 THE SINC FILTER IS A GENERALIZATION OF THE FILTER USED BY SHEPP
 + LOGAN[2].
 THE COSINE FILTER IS A COMMON LOWPASS FILTER APPLIED TO THE
 RECONSTRUCTION PROBLEM.
 THE HAMMING FILTER IS THE GENERALIZED HAMMING FILTER, A SPECIAL
 CASE OF WHICH WAS USED BY CHESLER + RIEDERER[4].
 THE CUTOFF VALUE REPRESENTS THE FRACTION OF THE FOURIER SPACE
 POWER RELATIVE TO THE RAY SPACING TO INCLUDE IN THE RECONSTRUCTION
 A CUTOFF AT THE FOLDING FREQUENCY OF THE 2D FOURIER TRANSFORM
 CORRESPONDS TO CUTOFF=PINC/PIXSIZ.  WHEN "CUTOFF" IS GIVEN AS 0.0,
 IT IS SET TO 1.0.  WHEN "CUTOFF" IS NEGATIVE, IT IS SET TO
 "2.0*PRJNUM*PINC/(PI*NELEM*PIXSIZ)".
 THE BACK PROJECTION REGION SIZE MAY BE USER SPECIFIED AS
 ANY EVEN NUMBER GREATER THAN NELEM (WITHIN REASON), OR IF NOT
 GIVEN DEFAULTS TO NELEM+1.
 EXAMPLE CONTROL CARD SEQUENCE
 EXECUTE RFL
 BAND LIMITING CUTOFF = -1.0 INTERPOLATION = 2
 WARNING.....
 THIS VERSION OF THE RFL METHOD WEIGHTS PROJECTION "NP"
 BY "(THETA(NP+1)-THETA(NP-1))/2.0".  IN ORDER TO ACCOMPLISH THIS
 IT IS ASSUMED THAT THE PROJECTIONS ARE ARRANGED IN INCREASING
 ORDER ON THETA AND THAT THETA(PRJNUM)-THETA(1) IS LESS THAN
 PI IN THE PARALLEL CASE OR TWOPI IN THE DIVERGENT CASE.
 REFERENCES:
 [1]  BRACEWELL + RIDDLE, INVERSION OF FAN-BEAM SCANS IN RADIO
 ASTROMOMY, ASTROPHYS. J., V.150, PP.427-434 (1967).
 [2]  RAMACHANDRAN + LAKSHMINARAYANAN, THREE-DIMENSIONAL
 RECONSTRUCTION FROM RADIOGRAPHS AND ELECTRON MICROGRAPHS:
 APPLICATION OF CONVOLUTIONS INSTEAD OF FOURIER TRANSFORMS,
 PROC. NAT. ACAD. SCI. U.S., V.68, PP.2236-2240 (1971).
 [3]  SHEPP + LOGAN, THE FOURIER RECONSTRUCTION OF A HEAD SECTION,
 IEEE TRANS. NUCL. SCI., NS-21, PP.21-43 (1974).
 [4]  CHESLER + RIEDERER, RIPPLE SUPPRESSION DURING RECONSTRUCTION
 IN TRANSVERSE TOMOGRAPHY, PHYS. MED. BIOL., V.20, PP.632-636
 (1975).
 */

#include <cstdio>
#include <cmath>

#include "blkdta.h"
#include "geom.h"
#include "spctrm.h"
#include "fourie.h"
#include "consts.h"
#include "uiod.h"

#include "int2str.h"
#include "rtfort.h"
#include "bckprj.h"
#include "qfilt.h"
#include "anglst.h"

#include "projfile.h"
#include "infile.h"

#include "rfl.h"

BOOLEAN rfl_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{
	REAL* reco = NULL;   //wei, 3/2005
	REAL* g;
	INTEGER gnum;

	INTEGER word;

	REAL thmod;
	REAL thfac;
	INTEGER melen;
	INTEGER narea;
	INTEGER nn;
	INTEGER ndelta;
	REAL theta;
	REAL sinth;
	REAL costh;
	REAL theta0;
	INTEGER np;
	REAL theta1;
	INTEGER np2;
	REAL theta2;
	REAL sinth2;
	REAL costh2;
	REAL w;
	INTEGER nr;
	REAL topinc;
	REAL scale;
	INTEGER nrow;
	INTEGER ncol;
	INTEGER ind;
	INTEGER i;
	INTEGER ip;
	INTEGER j;
	INTEGER jp;
	REAL r;
	REAL factr;
	REAL sum;
	REAL diff;
	INTEGER index;
	INTEGER offset;

	BOOLEAN eol;
	INTEGER n[3], filter;

	// modified calls to getwrd() to use the 4-parameter version. Lajos, Dec 16, 2004
	static const INTEGER fname[4] =
	{
			CHAR2INT('b', 'a', 'n', 'd'), CHAR2INT('s', 'i', 'n', 'c'),
			CHAR2INT('c', 'o', 's', 'i'), CHAR2INT('h', 'a', 'm', 'm') };

	// INITIALIZATION

	///xalg6=.false.
	if (iter > 1)
		return FALSE;
	thmod = Consts.pi;
	thfac = 2.0;
	if (!GeoPar.par)
	{
		thmod = Consts.twopi;
		thfac = 4.0;
	}
	word = InFile.getwrd(TRUE, &eol, fname, 4);

	for (filter = 0; filter < 4; filter++)
	{
		if (word == fname[filter])
			goto L30;
	}

	fprintf(output, "\n***** %4s is not a valid filter name", int2str(word));
	return TRUE;

	L30: Fourie.cutoff = InFile.getnum(FALSE, &eol);
	Fourie.interp = InFile.getint(FALSE, &eol);

	if (eol)
	{
		fprintf(output, "\n *** not enough parameters");
		return TRUE;
	}

	melen = InFile.getint(FALSE, &eol);
	if (Fourie.cutoff < 0.0)
		Fourie.cutoff = (REAL) 2.0 * GeoPar.prjnum * GeoPar.pinc / ((REAL) Consts.pi * GeoPar.nelem * GeoPar.pixsiz);
	if (Fourie.cutoff <= Consts.zero)
		Fourie.cutoff = 1.0;
	if (Fourie.cutoff > 1.0)
		Fourie.cutoff = 1.0;
	fprintf(output, "\n           rho-filtered layergram reconstruction algorithm");
	fprintf(output, "\n           using filter %4s", int2str(word));
	fprintf(output, "\n           cutoff (or alpha) = %9.4f", Fourie.cutoff);

	if ((Fourie.interp < -1) || (Fourie.interp > 6))
	{
		fprintf(output, "\n invalid interpolation %4i", Fourie.interp);
		return TRUE;
	}

	if (Fourie.interp > 0)
	{
		fprintf(output, "\n         %2i point lagrange interpolation", Fourie.interp);
	}
	if (Fourie.interp < 0)
	{
		fprintf(output, "\n           cubic spline interpolation");
	}
	if (Fourie.interp == 0)
	{
		fprintf(output, "\n           band limiting (sinc) interpolation");
	}
	if (!(melen > 0))
	{
		melen = GeoPar.nelem + 1;
	}

	if (melen < GeoPar.nelem)
	{
		fprintf(output, "\n***** melen is less than nelem, melen= %4i", melen);
		return TRUE;
	}

	fprintf(output, "\n           backprojection is performed on a %5i x %5i region", melen, melen);

	n[0] = melen;
	n[1] = melen / 2;
	n[2] = 1;

	if ((2 * n[1]) != melen)
	{
		fprintf(output, "\n***** melen must be even, melen= %4i", melen);
		return TRUE;
	}

	narea = (melen + 2) * melen;

	reco = new REAL[narea];

	for (nn = 0; nn < narea; nn++)
	{
		reco[nn] = 0.0;
	}

	// G IS THE BASE OF PROJECTION DATA

	ndelta = 2 * Fourie.interp;
	if (Fourie.interp < 0)
		ndelta = 6;
	gnum = GeoPar.numray(melen, GeoPar.pixsiz) + ndelta;

	g = new REAL[gnum];

	// BEGIN THE RECONSTRUCTION

	Anglst.getang(GeoPar.prjnum - 1, &theta, &sinth, &costh);

	theta0 = theta - thmod;

	for (np = 0; np < GeoPar.prjnum; np++)
	{

		ProjFile.ReadProj(np, g, gnum);

		Anglst.getang(np, &theta1, &sinth, &costh);

		// COMPUTE THE WEIGHT FOR THE PROJECTION

		np2 = np + 1;

		if (np == (GeoPar.prjnum - 1))
			np2 = 0;

		Anglst.getang(np2, &theta2, &sinth2, &costh2);

		if (np == (GeoPar.prjnum - 1))
			theta2 += thmod;

		w = (theta2 - theta0) / thfac;

		if (w <= 0.0)
		{
			fprintf(output, "\n *** angles are not in increasing order");
			if (reco != NULL)
				delete[] reco;   //wei 3/2005
			return TRUE;
		}

		Fourie.rinc = GeoPar.pinc;
		if (GeoPar.div)
			Fourie.rinc *= GeoPar.radius / GeoPar.stod;
		if (GeoPar.vri)
			Fourie.rinc = (REAL) (Fourie.rinc * MAX0(fabs(sinth), fabs(costh)));
		if (GeoPar.strip)
			w /= Fourie.rinc;
		theta0 = theta1;

		for (nr = 0; nr < gnum; nr++)
		{
			g[nr] *= w;
		}
		// BACK PROJECT THE DATA
		bckprj(reco, melen, g, gnum, sinth, costh, GeoPar.pixsiz / Fourie.rinc, Fourie.interp);

	}

	// COMPUTE 2-D FOURIER TRANSFORM
	delete[] g;  // bug 92 - Lajos - 03/02/2005

	rtfort(reco, n, 1);

	// MULTIPLY BY FILTER
	topinc = (REAL) (2.0 * GeoPar.pinc);

	scale = topinc / (melen * GeoPar.pixsiz);

	nrow = melen / 2 + 1;
	ncol = melen;
	ind = -2;

	for (i = 0; i < ncol; i++)
	{
		ip = MIN0(i, melen - i);
		for (j = 0; j < nrow; j++)
		{
			jp = j;
			ind += 2;
			r = (REAL) (sqrt(((REAL) (ip * ip + jp * jp))) * scale);
			factr = qfilt(r, Fourie.cutoff, filter) / topinc;

			reco[ind] *= factr;
			reco[ind + 1] *= factr;
		}
	}

	// INVERSE FOURIER TRANSFORM
	rtfort(reco, n, -1);

	// SCALE TO ESTIMATED AVERAGE DENSITY
	sum = 0.0;
	offset = (melen - 1 - GeoPar.nelem) / 2;

	ind = 0;
	for (i = 0; i < GeoPar.nelem; i++)
	{
		index = (i + offset) * melen + offset;
		for (j = 0; j < GeoPar.nelem; j++)
		{
			recon[ind] = reco[index];
			sum += reco[index];
			ind++;
			index++;
		}
	}

	diff = GeoPar.aveden - sum / GeoPar.area;

	for (i = 0; i < GeoPar.area; i++)
	{
		recon[i] += diff;
	}

	if (reco != NULL)
		delete[] reco;   //wei,3/2005
		return FALSE;
}

