/** @file mode.cpp
    @package snark14Input
    @author Bruno M. Carvalho and Deniz Sarioz
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
*/

#include "mode.h"
#include "misc.h"
#include "variables.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <qcheckbox.h>
#include <qframe.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qlayout.h>
#include <qvariant.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qmessagebox.h>

/** 
 *  Constructs a modewindow which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
@param void
@author Bruno M. Carvalho
@version 1.0 */
modewindow::modewindow( QWidget* parent,  const char* name, bool modal, WFlags fl )
    : QDialog( parent, name, modal, fl )
{
    if ( !name )
	setName( "modewindow" );
    resize( 400, 250 ); 
    setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, sizePolicy().hasHeightForWidth() ) );
    setMinimumSize( QSize( 400, 250 ) );
    setMaximumSize( QSize( 400, 250 ) );
    setCaption( tr( "Mode Window" ) );

    modelabel = new QLabel( this, "modelabel" );
    modelabel->setGeometry( QRect( 10, 10, 70, 30 ) ); 
    modelabel->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, modelabel->sizePolicy().hasHeightForWidth() ) );
    modelabel->setMinimumSize( QSize( 70, 30 ) );
    modelabel->setMaximumSize( QSize( 70, 30 ) );
    modelabel->setText( tr( "MODE" ) );

    lowerbox = new QCheckBox( this, "lowerbox" );
    lowerbox->setGeometry( QRect( 30, 60, 80, 30 ) ); 
    lowerbox->setText( tr( "LOWER" ) ); 
    QObject::connect(lowerbox,SIGNAL(clicked()),this,SLOT(activateLowerFields()));

    lowervalue = new QLineEdit( this, "lowervalue" );
    lowervalue->setGeometry( QRect( 130, 60, 110, 30 ) ); 
    lowervalue->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, lowervalue->sizePolicy().hasHeightForWidth() ) );
    lowervalue->setEnabled(FALSE);

    lowervaluelabel = new QLabel( this, "lowervaluelabel" );
    lowervaluelabel->setEnabled( FALSE );
    lowervaluelabel->setGeometry( QRect( 140, 40, 90, 20 ) ); 
    lowervaluelabel->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)1, (QSizePolicy::SizeType)1, lowervaluelabel->sizePolicy().hasHeightForWidth() ) );
    lowervaluelabel->setText( tr( "lower_value" ) );
    lowervaluelabel->setEnabled(FALSE);

    upperbox = new QCheckBox( this, "upperbox" );
    upperbox->setGeometry( QRect( 30, 130, 80, 30 ) ); 
    upperbox->setText( tr( "UPPER" ) ); 
    QObject::connect(upperbox,SIGNAL(clicked()),this,SLOT(activateUpperFields()));

    uppervalue = new QLineEdit( this, "uppervalue" );
    uppervalue->setGeometry( QRect( 130, 130, 110, 30 ) ); 
    uppervalue->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, uppervalue->sizePolicy().hasHeightForWidth() ) );
    uppervalue->setEnabled(FALSE);

    uppervaluelabel = new QLabel( this, "uppervaluelabel" );
    uppervaluelabel->setGeometry( QRect( 140, 110, 90, 20 ) ); 
    uppervaluelabel->setText( tr( "upper_value" ) ); 
    uppervaluelabel->setEnabled(FALSE);

    Line9 = new QFrame( this, "Line9" );
    Line9->setGeometry( QRect( -10, 175, 420, 16 ) ); 
    Line9->setFrameStyle( QFrame::HLine | QFrame::Sunken );

    okbutton = new QPushButton( this, "okbutton" );
    okbutton->setGeometry( QRect( 50, 190, 100, 40 ) ); 
    okbutton->setText( tr( "OK" ) );
    QObject::connect(okbutton,SIGNAL(clicked()),this,SLOT(checkValues()));

    cancelbutton = new QPushButton( this, "cancelbutton" );
    cancelbutton->setGeometry( QRect( 250, 190, 100, 40 ) ); 
    cancelbutton->setText( tr( "Cancel" ) );
    QObject::connect(cancelbutton,SIGNAL(clicked()),this,SLOT(reject()));
}

/**  
 *  Destroys the object and frees any allocated resources
 */
modewindow::~modewindow()
{
    // no need to delete child widgets, Qt does it all for us
}

/* Activates lower and upper fields 
@param void
@author Bruno M. Carvalho
@version 1.0 */
/* void modewindow::activateFields()
{
  if(modelabel->isChecked()) {
    lowerbox->setEnabled(TRUE);
    upperbox->setEnabled(TRUE);
  }
  else {
    lowerbox->setEnabled(FALSE);
    upperbox->setEnabled(FALSE);
    lowervalue->setEnabled(FALSE);
    lowervaluelabel->setEnabled(FALSE);
    uppervalue->setEnabled(FALSE);
    uppervaluelabel->setEnabled(FALSE);
  }
} */

/** Activates lower subfields 
@param void
@author Bruno M. Carvalho
@version 1.0 */
void modewindow::activateLowerFields()
{
  if(lowerbox->isChecked()) {
    lowervalue->setEnabled(TRUE);
    lowervaluelabel->setEnabled(TRUE);
  }
  else {
    lowervalue->setEnabled(FALSE);
    lowervaluelabel->setEnabled(FALSE);
  }
}

/** Activates upper subfields 
@param void
@author Bruno M. Carvalho
@version 1.0 */
void modewindow::activateUpperFields()
{
  if(upperbox->isChecked()) {
    uppervalue->setEnabled(TRUE);
    uppervaluelabel->setEnabled(TRUE);
  }
  else {
    uppervalue->setEnabled(FALSE);
    uppervaluelabel->setEnabled(FALSE);
  }
}

/** Returns a string for MODE input sequence
@param void
@return mode
@author Bruno M. Carvalho
@version 1.0 */
QString modewindow::getOutput()
{
  QString output;

  output.sprintf("MODE ");
  if(lowerbox->isChecked()) {
    output.append("LOWER ");
    output+=lowervalue->text();
    output.append(" ");
  }
  if(upperbox->isChecked()) {
    output.append("UPPER ");
    output+=uppervalue->text();
  }

  return output;
}

void modewindow::checkValues()
{
  int f=0;
  float l=0,u=0;
  
  if(lowerbox->isChecked()) {
    l=strtof((const char *)lowervalue->text());
    if(errnum) 
      f=1;
    if(upperbox->isChecked()) {
      u=strtof((const char *)uppervalue->text());
      if(errnum) 
        f=2;
      if(l>u && !f) 
        f=3;
    }
  }
  switch(f) {
  case 0: accept();
    break;
  case 1: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "Invalid lower_value format.");
  lowerbox->setFocus();
    break;
  case 2: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "Invalid upper_value format.");
  upperbox->setFocus();
    break;
  case 3: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "lower_value bigger than upper_value.");
  lowerbox->setFocus();
    break;
  }
}
